import json
from base64 import b64decode

from django.test.utils import override_settings
from django.contrib.auth.models import User
from django.core.urlresolvers import reverse, resolve
from django.test import TestCase

from s3direct import widgets


HTML_OUTPUT = (
    '<div class="s3direct" data-policy-url="/get_upload_params/">'
    '  <a class="file-link" target="_blank" href=""></a>'
    '  <a class="file-remove" href="#remove">Remove</a>'
    '  <input class="file-url" type="hidden" value="" id="None" name="filename" />'
    '  <input class="file-dest" type="hidden" value="foo">'
    '  <input class="file-input" type="file" />'
    '  <div class="progress progress-striped active">'
    '    <div class="bar"></div>'
    '  </div>'
    '</div>'
)

FOO_RESPONSE = {
    u'x-amz-algorithm': u'AWS4-HMAC-SHA256',
    u'form_action': u'https://s3.amazonaws.com/test-bucket',
    u'success_action_status': 201,
    u'acl': u'public-read',
    u'key': u'uploads/imgs/${filename}',
    u'content-type': u'image/jpeg'
}


class WidgetTestCase(TestCase):
    """
    This allows us to have 2 version of the same tests but with different settings.
    As opposed to inheriting test methods as doing that makes the failure stack hard to parse.
    TODO: Get rid of this base class and the appropriate subclass when positional setting support is dropped. See #48
    """
    
    def setUp(self):
        admin = User.objects.create_superuser('admin', 'u@email.com', 'admin')
        admin.save()
        
    def check_urls(self):
        reversed_url = reverse('s3direct')
        resolved_url = resolve('/get_upload_params/')
        self.assertEqual(reversed_url, '/get_upload_params/')
        self.assertEqual(resolved_url.view_name, 's3direct')
        
    def check_widget_html(self):
        widget = widgets.S3DirectWidget(dest='foo')
        self.assertEqual(widget.render('filename', None), HTML_OUTPUT)

    def check_signing_logged_in(self):
        self.client.login(username='admin', password='admin')
        data = {'dest': 'files', 'name': 'image.jpg', 'type': 'image/jpeg'}
        response = self.client.post(reverse('s3direct'), data)
        self.assertEqual(response.status_code, 200)

    def check_signing_logged_out(self):
        data = {'dest': 'files', 'name': 'image.jpg', 'type': 'image/jpeg'}
        response = self.client.post(reverse('s3direct'), data)
        self.assertEqual(response.status_code, 403)

    def check_allowed_type(self):
        data = {'dest': 'imgs', 'name': 'image.jpg', 'type': 'image/jpeg'}
        response = self.client.post(reverse('s3direct'), data)
        self.assertEqual(response.status_code, 200)

    def check_disallowed_type(self):
        data = {'dest': 'imgs', 'name': 'image.mp4', 'type': 'video/mp4'}
        response = self.client.post(reverse('s3direct'), data)
        self.assertEqual(response.status_code, 400)

    def check_allowed_type_logged_in(self):
        self.client.login(username='admin', password='admin')
        data = {'dest': 'vids', 'name': 'video.mp4', 'type': 'video/mp4'}
        response = self.client.post(reverse('s3direct'), data)
        self.assertEqual(response.status_code, 200)

    def check_disallowed_type_logged_out(self):
        data = {u'dest': u'vids', u'name': u'video.mp4', u'type': u'video/mp4'}
        response = self.client.post(reverse('s3direct'), data)
        self.assertEqual(response.status_code, 403)

    def check_signing_fields(self):
        self.client.login(username='admin', password='admin')
        data = {u'dest': u'imgs', u'name': u'image.jpg', u'type': u'image/jpeg'}
        response = self.client.post(reverse('s3direct'), data)
        response_dict = json.loads(response.content.decode())
        self.assertTrue(u'x-amz-signature' in response_dict)
        self.assertTrue(u'x-amz-credential' in response_dict)
        self.assertTrue(u'policy' in response_dict)
        self.assertDictContainsSubset(FOO_RESPONSE, response_dict)

    def check_signing_fields_unique_filename(self):
        data = {u'dest': u'misc', u'name': u'image.jpg', u'type': u'image/jpeg'}
        response = self.client.post(reverse('s3direct'), data)
        response_dict = json.loads(response.content.decode())
        self.assertTrue(u'x-amz-credential' in response_dict)
        self.assertTrue(u'x-amz-credential' in response_dict)
        self.assertTrue(u'policy' in response_dict)
        changed = FOO_RESPONSE.copy()
        changed['key'] = 'images/unique.jpg'
        self.assertDictContainsSubset(changed, response_dict)

    def check_policy_conditions(self):
        self.client.login(username='admin', password='admin')
        data = {u'dest': u'cached', u'name': u'video.mp4', u'type': u'video/mp4'}
        response = self.client.post(reverse('s3direct'), data)
        self.assertEqual(response.status_code, 200)

        response_dict = json.loads(response.content.decode())
        self.assertTrue('policy' in response_dict)
        policy_dict = json.loads(b64decode(response_dict['policy']).decode('utf-8'))
        self.assertTrue('conditions' in policy_dict)
        conditions_dict = policy_dict['conditions']
        self.assertEqual(conditions_dict[0]['bucket'], u'astoragebucketname')
        self.assertEqual(conditions_dict[1]['acl'], u'authenticated-read')
        self.assertEqual(conditions_dict[8]['Cache-Control'], u'max-age=2592000')
        self.assertEqual(conditions_dict[9]['Content-Disposition'], u'attachment')


@override_settings(S3DIRECT_DESTINATIONS={
    'misc': (lambda original_filename: 'images/unique.jpg',),
    'files': ('uploads/files', lambda u: u.is_staff,),
    'imgs': ('uploads/imgs', lambda u: True, ['image/jpeg', 'image/png'],),
    'vids': ('uploads/vids', lambda u: u.is_authenticated(), ['video/mp4'],),
    'cached': ('uploads/vids', lambda u: u.is_authenticated(), '*', 'authenticated-read',
               'astoragebucketname', 'max-age=2592000', 'attachment',),
})
class OldStyleSettingsWidgetTest(WidgetTestCase):
    """
    Test coverage for the older "positional" style of specifying settings.
    TODO: Remove me when positional settings are no longer supported.
    """

    def setUp(self):
        super(OldStyleSettingsWidgetTest, self).setUp()

    def test_urls(self):
        self.check_urls()

    def test_widget_html(self):
        self.check_widget_html()

    def test_signing_logged_in(self):
        self.check_signing_logged_in()

    def test_signing_logged_out(self):
        self.check_signing_logged_out()

    def test_allowed_type(self):
        self.check_allowed_type()

    def test_disallowed_type(self):
        self.check_disallowed_type()

    def test_allowed_type_logged_in(self):
        self.check_allowed_type_logged_in()

    def test_disallowed_type_logged_out(self):
        self.check_disallowed_type_logged_out()

    def test_signing_fields(self):
        self.check_signing_fields()

    def test_signing_fields_unique_filename(self):
        self.check_signing_fields_unique_filename()

    def test_policy_conditions(self):
        self.check_policy_conditions()


class WidgetTest(WidgetTestCase):
    
    def setUp(self):
        super(WidgetTest, self).setUp()

    def test_urls(self):
        self.check_urls()

    def test_widget_html(self):
        self.check_widget_html()

    def test_signing_logged_in(self):
        self.check_signing_logged_in()

    def test_signing_logged_out(self):
        self.check_signing_logged_out()

    def test_allowed_type(self):
        self.check_allowed_type()

    def test_disallowed_type(self):
        self.check_disallowed_type()

    def test_allowed_type_logged_in(self):
        self.check_allowed_type_logged_in()

    def test_disallowed_type_logged_out(self):
        self.check_disallowed_type_logged_out()

    def test_signing_fields(self):
        self.check_signing_fields()

    def test_signing_fields_unique_filename(self):
        self.check_signing_fields_unique_filename()

    def test_policy_conditions(self):
        self.check_policy_conditions()



import json

from django.http import HttpResponse
from django.views.decorators.http import require_POST

from .utils import create_upload_data, get_s3direct_destinations


@require_POST
def get_upload_params(request):
    content_type = request.POST['type']
    filename = request.POST['name']

    dest = get_s3direct_destinations().get(request.POST['dest'])

    if not dest:
        data = json.dumps({'error': 'File destination does not exist.'})
        return HttpResponse(data, content_type="application/json", status=400)

    key = dest.get('key')
    auth = dest.get('auth')
    allowed = dest.get('allowed')
    acl = dest.get('acl')
    bucket = dest.get('bucket')
    cache_control = dest.get('cache_control')
    content_disposition = dest.get('content_disposition')

    if not acl:
        acl = 'public-read'

    if not key:
        data = json.dumps({'error': 'Missing destination path.'})
        return HttpResponse(data, content_type="application/json", status=403)

    if auth and not auth(request.user):
        data = json.dumps({'error': 'Permission denied.'})
        return HttpResponse(data, content_type="application/json", status=403)

    if (allowed and content_type not in allowed) and allowed != '*':
        data = json.dumps({'error': 'Invalid file type (%s).' % content_type})
        return HttpResponse(data, content_type="application/json", status=400)

    if hasattr(key, '__call__'):
        key = key(filename)
    elif key == '/':
        key = '${filename}'
    else:
        # The literal string '${filename}' is an S3 field variable for key.
        # https://aws.amazon.com/articles/1434#aws-table
        key = '%s/${filename}' % key

    data = create_upload_data(
        content_type, key, acl, bucket, cache_control, content_disposition)

    return HttpResponse(json.dumps(data), content_type="application/json")



import hashlib
import hmac
import json
from datetime import datetime, timedelta
from base64 import b64encode

from django.conf import settings


REGIONS = {
    'us-east-1': 's3.amazonaws.com',
    'us-west-2': 's3-us-west-2.amazonaws.com',
    'us-west-1': 's3-us-west-1.amazonaws.com',
    'eu-west-1': 's3-eu-west-1.amazonaws.com',
    'eu-central-1': 's3.eu-central-1.amazonaws.com',
    'ap-southeast-1': 's3-ap-southeast-1.amazonaws.com',
    'ap-southeast-2': 's3-ap-southeast-2.amazonaws.com',
    'ap-northeast-1': 's3-ap-northeast-1.amazonaws.com',
    'sa-east-1': 's3-sa-east-1.amazonaws.com',
}


def get_at(index, t):
    try:
        value = t[index]
    except IndexError:
        value = None
    return value


# NOTE: Don't use constant as it will break ability to change at runtime (E.g. tests)
def get_s3direct_destinations():
    """Returns s3direct destinations, converting old format if necessary."""
    destinations = getattr(settings, 'S3DIRECT_DESTINATIONS', None)
    if destinations is None:
        return None

    # TODO: Remove when older "positional" style settings are no longer supported
    converted_destinations = {}
    key_mapping = {
        0: 'key',
        1: 'auth',
        2: 'allowed',
        3: 'acl',
        4: 'bucket',
        5: 'cache_control',
        6: 'content_disposition',
    }
    if destinations:
        for dest, dest_value in destinations.items():
            if type(dest_value) is tuple or type(dest_value) is list:
                converted_destinations[dest] = {}
                for index, key_name in key_mapping.items():
                    converted_destinations[dest][key_name] = get_at(index, dest_value)
            else:
                converted_destinations[dest] = dest_value

    return converted_destinations


def create_upload_data(content_type, key, acl, bucket=None, cache_control=None,
                       content_disposition=None):
    access_key = settings.AWS_ACCESS_KEY_ID
    secret_access_key = settings.AWS_SECRET_ACCESS_KEY
    bucket = bucket or settings.AWS_STORAGE_BUCKET_NAME
    region = getattr(settings, 'S3DIRECT_REGION', None)
    endpoint = REGIONS.get(region, 's3.amazonaws.com')

    expires_in = datetime.utcnow() + timedelta(seconds=60*5)
    expires = expires_in.strftime('%Y-%m-%dT%H:%M:%S.000Z')
    now_date = datetime.utcnow().strftime('%Y%m%dT%H%M%S000Z')
    raw_date = datetime.utcnow().strftime('%Y%m%d')

    policy_dict = {
            "expiration": expires,
            "conditions": [
                {"bucket": bucket},
                {"acl": acl},
                ["starts-with", "$key", ''],
                {"success_action_status": '201'},
                {"x-amz-credential": '%s/%s/%s/s3/aws4_request' % (
                    access_key,
                    raw_date, region
                )},
                {"x-amz-algorithm": "AWS4-HMAC-SHA256"},
                {"x-amz-date": now_date},
                {"content-type": content_type},
            ]
        }

    if cache_control:
        policy_dict['conditions'].append({'Cache-Control': cache_control})

    if content_disposition:
        policy_dict['conditions'].append({
            'Content-Disposition': content_disposition
        })

    policy_object = json.dumps(policy_dict)

    policy = b64encode(
        policy_object.replace('\n', '').replace('\r', '').encode())

    date_key = hmac.new(b'AWS4' + secret_access_key.encode('utf-8'),
                        msg=raw_date.encode('utf-8'),
                        digestmod=hashlib.sha256).digest()

    date_region_key = hmac.new(date_key, msg=region.encode('utf-8'),
                               digestmod=hashlib.sha256).digest()

    date_region_service_key = hmac.new(date_region_key, msg=b's3',
                                       digestmod=hashlib.sha256).digest()

    signing_key = hmac.new(date_region_service_key, msg=b'aws4_request',
                           digestmod=hashlib.sha256).digest()

    signature = hmac.new(signing_key, msg=policy,
                         digestmod=hashlib.sha256).hexdigest()

    structure = getattr(settings, 'S3DIRECT_URL_STRUCTURE', 'https://{0}/{1}')
    bucket_url = structure.format(endpoint, bucket)

    return_dict = {
        # FIXME: .decode() does nothing, b64decode works but is decoding really intended?
        "policy": policy.decode(),
        "success_action_status": 201,
        "x-amz-credential": "%s/%s/%s/s3/aws4_request" % (
            access_key, raw_date, region
        ),
        "x-amz-date": now_date,
        "x-amz-signature": signature,
        "x-amz-algorithm": "AWS4-HMAC-SHA256",
        "form_action": bucket_url,
        "key": key,
        "acl": acl,
        "content-type": content_type
    }

    if cache_control:
        return_dict['Cache-Control'] = cache_control

    if content_disposition:
        return_dict['Content-Disposition'] = content_disposition

    return return_dict
